def dataframe_from_result_table(table, raise_errors=True):
    import pandas
    import json

    kusto_to_dataframe_data_types = {
        "bool": "bool",
        "uint8": "int64",
        "int16": "int64",
        "uint16": "int64",
        "int": "int64",
        "uint": "int64",
        "long": "int64",
        "ulong": "int64",
        "float": "float64",
        "real": "float64",
        "decimal": "float64",
        "string": "object",
        "datetime": "datetime64[ns]",
        "guid": "object",
        "timespan": "timedelta64[ns]",
        "dynamic": "object",
        # Support V1
        "DateTime": "datetime64[ns]",
        "Int32": "int32",
        "Int64": "int64",
        "Double": "float64",
        "String": "object",
        "SByte": "object",
        "Guid": "object",
        "TimeSpan": "object",
    }

    """Returns Pandas data frame."""
    if not table.columns or not table.rows:
        return pandas.DataFrame()

    frame = pandas.DataFrame(table.rows, columns=[column.column_name for column in table.columns])

    for column in table.columns:
        col_name = column.column_name
        col_type = column.column_type
        if col_type.lower() == "timespan":
            frame[col_name] = pandas.to_timedelta(
                frame[col_name].apply(lambda t: t.replace(".", " days ") if t and "." in t.split(":")[0] else t)
            )
        elif col_type.lower() == "dynamic":
            frame[col_name] = frame[col_name].apply(lambda x: json.loads(x) if x else None)
        elif col_type in kusto_to_dataframe_data_types:
            pandas_type = kusto_to_dataframe_data_types[col_type]
            frame[col_name] = frame[col_name].astype(pandas_type, errors="raise" if raise_errors else "ignore")

    return frame



"""All kusto exceptions that can be thrown from this client."""


class KustoError(Exception):
    """Base class for all exceptions raised by the Kusto Python Client Libraries."""

    pass


class KustoServiceError(KustoError):
    """Raised when the Kusto service was unable to process a request."""

    def __init__(self, messages, http_response, kusto_response=None):
        super(KustoServiceError, self).__init__(self, messages)
        self.http_response = http_response
        self.kusto_response = kusto_response

    def get_raw_http_response(self):
        """Gets the http response."""
        return self.http_response

    def is_semantic_error(self):
        """Checks if a response is a semantic error."""
        return "Semantic error:" in self.http_response.text

    def has_partial_results(self):
        """Checks if a response exists."""
        return self.kusto_response is not None

    def get_partial_results(self):
        """Gets the Kusto response."""
        return self.kusto_response


class KustoClientError(KustoError):
    """Raised when a Kusto client is unable to send or complete a request."""

    pass



import json
import six
from enum import Enum
from . import _converters


class WellKnownDataSet(Enum):
    """Categorizes data tables according to the role they play in the data set that a Kusto query returns."""

    PrimaryResult = "PrimaryResult"
    QueryCompletionInformation = "QueryCompletionInformation"
    TableOfContents = "TableOfContents"
    QueryProperties = "QueryProperties"


class KustoResultRow(object):
    """Iterator over a Kusto result row."""

    def __init__(self, columns, row):
        self.columns = columns
        self.row = row

        # Here we keep converter functions for each type that we need to take special care
        # (e.g. convert)
        self.convertion_funcs = {
            "datetime": _converters.to_datetime,
            "timespan": _converters.to_timedelta,
            "DateTime": _converters.to_datetime,
            "TimeSpan": _converters.to_timedelta,
        }

    @property
    def columns_count(self):
        return len(self.columns)

    def __iter__(self):
        for i in range(self.columns_count):
            yield self[i]

    def __getitem__(self, key):
        if isinstance(key, six.integer_types):
            column = self.columns[key]
            value = self.row[key]
        else:
            column = next((column for column in self.columns if column.column_name == key), None)
            if not column:
                raise LookupError(key)
            value = self.row[column.ordinal]
        if column.column_type in self.convertion_funcs:
            return self.convertion_funcs[column.column_type](value)
        return value

    def __len__(self):
        return self.columns_count

    def to_dict(self):
        return {c.column_name: self.row[c.ordinal] for c in self.columns}

    def __str__(self):
        return self.row

    def __repr__(self):
        return "KustoResultRow({},{})".format(self.columns, self.row)


class KustoResultColumn(object):
    def __init__(self, json_column, ordianl):
        self.column_name = json_column["ColumnName"]
        self.column_type = json_column.get("ColumnType") or json_column["DataType"]
        self.ordinal = ordianl

    def __repr__(self):
        return "KustoResultColumn({},{})".format(
            json.dumps({"ColumnName": self.column_name, "ColumnType": self.column_type}), self.ordinal
        )


class KustoResultTable(object):
    """Iterator over a Kusto result table."""

    def __init__(self, json_table):
        self.table_name = json_table.get("TableName")
        self.table_id = json_table.get("TableId")
        self.table_kind = WellKnownDataSet[json_table["TableKind"]] if "TableKind" in json_table else None
        self.columns = [KustoResultColumn(column, index) for index, column in enumerate(json_table["Columns"])]

        self.rows = json_table["Rows"]

    @property
    def rows_count(self):
        return len(self.rows)

    @property
    def columns_count(self):
        return len(self.columns)

    def __len__(self):
        return self.rows_count

    def __iter__(self):
        for row in self.rows:
            yield KustoResultRow(self.columns, row)

    def __getitem__(self, key):
        return KustoResultRow(self.columns, self.rows[key])

    def to_dict(self):
        return {"name": self.table_name, "kind": self.table_kind, "data": [r.to_dict() for r in self]}

    def __str__(self):
        return json.dumps(self.to_dict())
